const Spring = require('gravitas/src/Spring');

const InteractEvent  = require('./InteractEvent');
const Interaction    = require('./Interaction');
const modifiers      = require('./modifiers');
const utils          = require('./utils');
const animationFrame = require('./utils/raf');

Interaction.signals.on('new', function (interaction) {
  interaction.inertiaStatus = {
    active   : false,
    allowResume: false,
    ending   : false,

    startEvent: null,
    upCoords  : {},

    xe: 0, ye: 0,
    sx: 0, sy: 0,

    t0: 0,
    vx0: 0, vys: 0,
    duration: 0,

    lambda_v0: 0,
    one_ve_v0: 0,
    i  : null,

    modelX: new Spring(1, 200, 28.3),
    modelY: new Spring(1, 200, 28.3),
  };

  interaction.boundInertiaFrame = () => inertiaFrame  .apply(interaction);
});

Interaction.signals.on('down', function ({ interaction, event, pointer, eventTarget }) {
  const status = interaction.inertiaStatus;

  // Check if the down event hits the current inertia target
  if (status.active) {
    let element = eventTarget;

    // climb up the DOM tree from the event target
    while (utils.is.element(element)) {

      // if interaction element is the current inertia target element
      if (element === interaction.element) {
        // stop inertia
        animationFrame.cancel(status.i);
        status.active = false;
        interaction.simulation = null;

        // update pointers to the down event's coordinates
        interaction.updatePointer(pointer);
        utils.setCoords(interaction.curCoords, interaction.pointers);

        // fire appropriate signals
        const signalArg = { interaction };
        Interaction.signals.fire('before-action-move', signalArg);
        Interaction.signals.fire('resume'            , signalArg);

        // fire a reume event
        const resumeEvent = new InteractEvent(interaction,
                                              event,
                                              interaction.prepared.name,
                                              'inertiaresume',
                                              interaction.element);

        interaction.target.fire(resumeEvent);
        interaction.prevEvent = resumeEvent;
        modifiers.resetStatuses(interaction.modifierStatuses);

        utils.copyCoords(interaction.prevCoords, interaction.curCoords);
        break;
      }
      element = utils.parentElement(element);
    }
  }
});

Interaction.signals.on('up', function ({ interaction, event }) {
  const status = interaction.inertiaStatus;

  if (!interaction.interacting() || status.active) { return; }

  const target = interaction.target;
  const options = target && target.options;
  const inertiaOptions = options && interaction.prepared.name && options[interaction.prepared.name].inertia;

  const now = new Date().getTime();
  const statuses = {};
  const page = utils.extend({}, interaction.curCoords.page);
  const delayedRelease = (now - interaction.curCoords.timeStamp) >= 50;
  const velocity = delayedRelease
    ? { x: 0, y: 0 }
    : {
      x: interaction.pointerDelta.client.vx,
      y: interaction.pointerDelta.client.vy,
    };
  const speed = delayedRelease? 0 : interaction.pointerDelta.client.speed;
  let inertiaPossible = false;
  let inertia = false;
  let smoothEnd = false;
  let modifierResult;

  // check if inertia should be started
  inertiaPossible = (inertiaOptions && inertiaOptions.enabled
                     && interaction.prepared.name !== 'gesture'
                     && event !== status.startEvent);

  inertia = (inertiaPossible
            && !delayedRelease
            && speed > inertiaOptions.minSpeed
            && speed > inertiaOptions.endSpeed);

  // smoothEnd
  if (inertiaPossible && !inertia) {
    modifiers.resetStatuses(statuses);

    modifierResult = modifiers.setAll({ interaction, page, statuses, preEnd: true, requireEndOnly: true });

    if (modifierResult.shouldMove && modifierResult.locked) {
      inertia = true;
      smoothEnd = true;

      status.xe = modifierResult.dx;
      status.ye = modifierResult.dy;
    }
  }

  if (!inertia) { return; }

  utils.copyCoords(status.upCoords, interaction.curCoords);

  interaction.pointers[0] = status.startEvent =
    new InteractEvent(interaction, event, interaction.prepared.name, 'inertiastart', interaction.element);

  status.t0 = now;

  status.active = true;
  status.allowResume = inertiaOptions.allowResume;
  interaction.simulation = status;

  target.fire(status.startEvent);

  // < for old calcInertia
  status.vx0 = velocity.x;
  status.vy0 = velocity.y;
  status.v0 = speed;
  // >

  if (!smoothEnd) {
    calcInertia(interaction, status);

    modifiers.resetStatuses(statuses);

    page.x += status.xe;
    page.y += status.ye;

    modifierResult = modifiers.setAll({ interaction, page, statuses, preEnd: true, requireEndOnly: true });

    status.xe += modifierResult.dx;
    status.ye += modifierResult.dy;
  }

  const modelX = status.modelX;
  const modelY = status.modelY;

  modelX.snap(0);
  modelY.snap(0);
  modelX.setEnd(1, velocity.x / status.xe);
  modelY.setEnd(1, velocity.y / status.ye);

  status.sx = status.sy = 0;
  status.i = animationFrame.request(interaction.boundInertiaFrame);
});

Interaction.signals.on('stop-active', function ({ interaction }) {
  const status = interaction.inertiaStatus;

  if (status.active) {
    animationFrame.cancel(status.i);
    status.active = status.ending = false;
    interaction.simulation = null;
  }
});

function calcInertia (interaction, status) {
  const inertiaOptions = interaction.target.options[interaction.prepared.name].inertia;
  const lambda = inertiaOptions.resistance;
  const inertiaDur = -Math.log(inertiaOptions.endSpeed / status.v0) / lambda;

  status.x0 = interaction.prevEvent.pageX;
  status.y0 = interaction.prevEvent.pageY;
  status.t0 = status.startEvent.timeStamp / 1000;

  status.xe = (status.vx0 - inertiaDur) / lambda;
  status.ye = (status.vy0 - inertiaDur) / lambda;
  status.te = inertiaDur;

  status.lambda_v0 = lambda / status.v0;
  status.one_ve_v0 = 1 - inertiaOptions.endSpeed / status.v0;
}

function inertiaFrame () {
  updateInertiaCoords(this);
  utils.setCoordDeltas(this.pointerDelta, this.prevCoords, this.curCoords);

  const status = this.inertiaStatus;
  const done = status.modelX.done() && status.modelY.done();

  status.sx = status.modelX.x() * status.xe;
  status.sy = status.modelY.x() * status.ye;

  if (done) {
    status.ending = true;
  }

  this.doMove();

  if (done) {
    this.end(status.startEvent);
    status.active = status.ending = false;
    this.simulation = null;
  }
  else {
    status.i = animationFrame.request(this.boundInertiaFrame);
  }

  utils.copyCoords(this.prevCoords, this.curCoords);
}

function updateInertiaCoords (interaction) {
  // return if inertia isn't running
  if (!interaction.inertiaStatus.active) { return; }

  const pageUp   = interaction.inertiaStatus.upCoords.page;
  const clientUp = interaction.inertiaStatus.upCoords.client;

  utils.setCoords(interaction.curCoords, [ {
    pageX  : pageUp.x   + interaction.inertiaStatus.sx,
    pageY  : pageUp.y   + interaction.inertiaStatus.sy,
    clientX: clientUp.x + interaction.inertiaStatus.sx,
    clientY: clientUp.y + interaction.inertiaStatus.sy,
  } ]);
}
window.cmk = function getSrpingCMK (m, k, c) {// ({ _c: c, _m: m, _k: k }) {
  return c * c - 4 * m * k;
};
